#!/usr/bin/env python3

from setuptools import setup, find_packages, Extension
from setuptools.command.build_py import build_py
import os
import os.path
import subprocess

try:
	import numpy as np
	numpy_include_dirs = [np.get_include()]
except ModuleNotFoundError:
	numpy_include_dirs = []

class BuildCOSP(build_py):
	def run(self):
		subprocess.run('make', cwd='cosp', check=True)
		subprocess.run('make', check=True)
		build_py.run(self)

setup(
	name='alcf',
	version='1.8.0',
	description='Automatic Lidar and Ceilometer Framework (ALCF)',
	author='Peter Kuma, Adrian J. McDonald, Olaf Morgenstern, Richard Querel, Israel Silber, Connor J. Flynn',
	author_email='peter@peterkuma.net',
	license='MIT',
	entry_points={
		'console_scripts': 'alcf = alcf.bin.alcf:main_wrapper',
	},
	packages=find_packages(),
	ext_modules=[
		Extension(
			'alcf.algorithms.interp',
			['alcf/algorithms/interp.pyx'],
			include_dirs = numpy_include_dirs,
		),
	],
	zip_safe=False,
	package_data={'alcf': ['cosp_alcf'] + \
		[os.path.join('fonts', x) for x in os.listdir('alcf/fonts')]},
	data_files=[('share/man/man1',
		[os.path.join('man', x) for x in os.listdir('man')])],
	setup_requires=[
		'cython',
		'numpy',
	],
	install_requires=[
		'numpy>=1.24.2',
		'scipy>=1.10.1',
		'matplotlib>=3.7.2',
		'netCDF4>=1.6.4',
		'cl2nc>=3.4.0',
		'mpl2nc>=1.3.6',
		'aquarius-time>=0.4.0',
		'ds-format>=4.1.0',
		'pst-format>=2.0.0',
		'astropy>=5.3.2',
		'cftime>=1.6.2',
		'requests>=2.31.0',
		'cdsapi>=0.7.0',
	],
	keywords=['alc', 'ceilometer', 'lidar', 'atmosphere', 'model', 'simulator', 'nwp', 'gcm', 'cosp', 'actsim', 'vaisala', 'cl51', 'cl31', 'lufft', 'chm-15k', 'minimpl', 'amps', 'merra-2', 'um'],
	url='https://alcf.peterkuma.net',
	classifiers=[
		'Development Status :: 5 - Production/Stable',
		'Environment :: Console',
		'Intended Audience :: Science/Research',
		'License :: OSI Approved :: MIT License',
		'Operating System :: POSIX',
		'Programming Language :: Python :: 3',
		'Programming Language :: Cython',
		'Programming Language :: Fortran',
		'Topic :: Scientific/Engineering :: Atmospheric Science',
		'Topic :: Scientific/Engineering :: Physics',
		'Topic :: Scientific/Engineering :: Visualization',
	],
	cmdclass={
		'build_py': BuildCOSP,
	}
)



from subprocess import check_call, call
import sys
import os
import re
import glob
import tempfile
import shutil

TYPES = {
	'amps': {
		'ext': ['grb'],
		'outfile_is_directory': True,
		'cmdf': lambda infile, outfile: ['ncl_convert2nc', infile, '-o', outfile],
	},
	'cl31': {
		'ext': ['dat', 'DAT', 'asc', 'ASC'],
		'cmdf': lambda infile, outfile: ['cl2nc', infile, outfile],
	},
	'cl51': {
		'ext': ['dat', 'DAT', 'asc', 'ASC'],
		'cmdf': lambda infile, outfile: ['cl2nc', infile, outfile],
	},
	'jra55': {
		'ext': [None],
		'cmdf': lambda infile, outfile: ['grib_to_netcdf', '-o', outfile, infile],
	},
	'mpl': {
		'ext': ['mpl', 'MPL'],
		'cmdf': lambda infile, outfile: ['mpl2nc', infile, outfile],
	},
}

def find(input_, output, inext, outext, recursive=False):
	if not input_.endswith('/'):
		input_ += '/'
	if recursive:
		path = '**' if inext is None else '**/*.' + glob.escape(inext)
	else:
		path = '*' if inext is None else '*.' + glob.escape(inext)
	pattern = os.path.join(glob.escape(input_), path)
	files = sorted(glob.glob(pattern, recursive=recursive))
	input2 = []
	output2 = []
	for file_ in files:
		if not os.path.isfile(file_):
			continue
		if not file_.startswith(input_):
			continue
		infile = file_[len(input_):]
		if inext is not None and infile.endswith('.' + inext):
			outfile = infile[:-len(inext)] + outext
		else:
			outfile = infile + '.' + outext
		input2 += [os.path.join(input_, infile)]
		output2 += [os.path.join(output, outfile)]
	return input2, output2

def run(type_, input_, output, *args, **kwargs):
	"""
alcf-convert -- Convert input instrument or model data to NetCDF.
============

Synopsis
--------

    alcf convert [options] <type> [--] <input> <output>

Description
-----------

Arguments following `--` are treated as literal strings. Use this delimiter if the input or output file names might otherwise be interpreted as non-strings, e.g. purely numerical file names.

Arguments
---------

- `type`: Input data type (see Types below).
- `input`: Input filename or dirname. If `input` is a directory, all data files ending with the correct file extension (see Types below) in `input` are converted to corresponding `.nc` files in the directory `output`. If the option `-r` is supplied, the directory is processed recursively.
- `output`: Output filename or dirname.

Options
-------

- `-r`: Process the input directory recursively. The same directory structure is created under `output`.

Types
-----

- `amps`: Antarctic Mesoscale Prediction System (AMPS) GRIB files. Input file extension `.grb`.
- `cl31`: Vaisala CL31. Input file extension `.dat`, `.DAT`, `.asc` or `.ASC`.
- `cl51`: Vaisala CL51. Input file extension `.dat`, `.DAT`, `.asc` or `.ASC`.
- `jra55`: JRA-55 reanalysis. No input file extension.
- `mpl`: MiniMPL or MPL. Input file extension `.mpl` or `.MPL`.

Examples
--------

Convert raw Vaisala CL51 data in `cl51_dat` to NetCDF and store the output in the directory `cl51_nc`.

    alcf convert cl51 cl51_dat cl51_nc

Convert JRA-55 data in `jra55_grib` to NetCDF and store the output in the directory `jra55_nc`.

    alcf convert jra55 jra55_grib jra55_nc
	"""
	recursive = kwargs.get('r', False)

	if type_ not in TYPES:
		raise ValueError('Invalid type: %s' % type_)
	desc = TYPES[type_]

	if os.path.isdir(input_) and not os.path.isdir(output):
		raise IOError('%s: the output path does not exist or is not a directory' % output)

	if os.path.isfile(input_) and os.path.isdir(output):
		basename = os.path.basename(input_)
		for ext in desc['ext']:
			if ext is not None and basename.endswith('.' + ext):
				basename = basename[:-(len(ext) + 1)]
				break
		output = os.path.join(output, basename + '.nc')

	if os.path.isdir(input_):
		input2 = []
		output2 = []
		for ext in desc['ext']:
			in_, out = find(input_, output, ext, 'nc',
				recursive=recursive)
			input2 += in_
			output2 += out
	else:
		input2, output2 = [input_], [output]

	for infile, outfile in zip(input2, output2):
		if recursive:
			os.makedirs(os.path.dirname(outfile), exist_ok=True)
		if desc.get('outfile_is_directory'):
			with tempfile.TemporaryDirectory() as tmpdir:
				cmd = desc['cmdf'](infile, tmpdir)
				print(' '.join(cmd))
				call(cmd, stdout=sys.stdout, stderr=sys.stderr)
				name = os.path.splitext(os.path.basename(infile))[0]
				filename = os.path.join(tmpdir, name + '.nc')
				shutil.move(filename, outfile)
		else:
			cmd = desc['cmdf'](infile, outfile)
			print(' '.join(cmd))
			call(cmd, stdout=sys.stdout, stderr=sys.stderr)



import os
import ds_format as ds
import numpy as np
import aquarius_time as aq
from alcf.lidars import LIDARS

def read_time_periods(filename):
	tp = []
	with open(filename, 'r') as f:
		for line in f.readlines():
			if line.strip() == '':
				continue
			s1, s2 = line.split()
			start = aq.from_iso(s1)
			end = aq.from_iso(s2)
			tp.append([start, end])
	return tp

def run(type_, time_periods, input_, output, eta=0.7):
	"""
alcf-calibrate -- Calibrate ALC backscatter.
==============

Synopsis
--------

    alcf calibrate <type> [--] <time_periods> <input> <output>

Description
-----------

Calibration based the O'Connor et al. (2004) method of lidar ratio (LR) in fully opaque stratocumulus clouds.

Arguments following `--` are treated as literal strings. Use this delimiter if the input or output file names might otherwise be interpreted as non-strings, e.g. purely numerical file names.

Arguments
---------

- `type`: Lidar type (see Types below).
- `time_periods`: File containing calibration time periods of stratocumulus clouds in the backscatter profiles (see Time periods below).
- `input`: Input directory containing NetCDF files (the output of `alcf lidar`).
- `output`: Output calibration file.

Types
-----

- `chm15k`: Lufft CHM 15k.
- `cl31`: Vaisala CL31.
- `cl51`: Vaisala CL51.
- `minimpl`: Sigma Space MiniMPL.
- `mpl`: Sigma Space MPL.

Files
-----

Time periods file is a text file containting start and end time (see Time format below) of a time period separated by whitespace, one time period per line:

```
<start> <end>
<start> <end>
...
```

where `start` is the start time and `end` is the end time.

Time format
-----------

`YYYY-MM-DD[THH:MM[:SS]]`, where `YYYY` is year, `MM` is month, `DD` is day, `HH` is hour, `MM` is minute, `SS` is second. Example: `2000-01-01T00:00:00`.

Examples
--------

Read time periods from `time_periods.txt`, lidar profiles from the directory `lidar` and write the calibration coefficient to `calibration.txt`.

    alcf calibrate time_periods.txt lidar calibration.txt
	"""
	lidar = LIDARS.get(type_)
	tp = read_time_periods(time_periods)
	files = sorted(os.listdir(input_))
	lr = []
	for file_ in files:
		filename = os.path.join(input_, file_)
		print('<- %s' % filename)
		d = ds.read(filename, ['time'])
		mask = np.zeros(len(d['time']), dtype=bool)
		for period in tp:
			mask |= (d['time'] >= period[0]) & \
				(d['time'] < period[1])
		d = ds.read(filename, ['lr'], {'time': mask})
		lr.append(d['lr'])
	lr = np.hstack(lr)
	lr_median = np.median(lr)
	calibration_ceoff = lidar.CALIBRATION_COEFF*lr_median/lidar.SC_LR
	print('-> %s' % output)
	with open(output, 'w') as f:
		f.write('lidar: %s wavelength: %d calibration_coeff: %f lr_median: %f\n' % (
			type_,
			lidar.WAVELENGTH,
			calibration_ceoff,
			lr_median,
		))



#!/usr/bin/env python

import sys
import ds_format as ds
import numpy as np
from scipy.stats import norm
from scipy.optimize import minimize_scalar
from alcf import misc

def cloud_detection(d,
	cloud_threshold=2e-6,
	cloud_threshold_exp=None,
	cloud_nsd=5,
	bsd=None,
	bsd_z=8000,
	**options
):
	b = d['backscatter']
	if b.ndim == 3:
		n, m, l = b.shape
	else:
		n, m = b.shape

	if bsd is None:
		bsd2 = d.get('backscatter_sd')
	else:
		bsd2 = np.full(b.shape, bsd, np.float64)
		f = (d['zfull']/bsd_z)**2
		if b.ndim == 3:
			for i in range(b.shape[2]):
				bsd2[:,:,i] *= f
		else:
			bsd2 *= f
	bmol = d.get('backscatter_mol')
	x = np.copy(b)
	if bsd2 is not None:
		x -= cloud_nsd*bsd2
	if bmol is not None:
		if len(x.shape) == 3:
			for i in range(x.shape[2]):
				x[:,:,i] -= bmol
		else:
			x -= bmol

	cloud_mask = np.zeros(x.shape, np.byte)
	if cloud_threshold_exp is not None:
		ct_x, ct_y, ct_h = cloud_threshold_exp
		for i in range(n):
			alt = 0 if np.isnan(d['altitude'][i]) else d['altitude'][i]
			height = d['zfull'] - alt
			height[height < 0] = 0
			ct = ct_y + (ct_x - ct_y)*0.5**(height/ct_h)
			if x.ndim == 3:
				for k in range(l):
					cloud_mask[i,:,k] = x[i,:,k] >= ct
			else:
				cloud_mask[i,:] = x[i,:] >= ct
	else:
		cloud_mask[::] = x >= cloud_threshold

	d['cloud_mask'] = cloud_mask
	d['.']['cloud_mask'] = {
		'.dims': d['.']['backscatter']['.dims'],
		'long_name': 'cloud mask',
		'units': '1',
	}

def stream(dd, state, **options):
	return misc.stream(dd, state, cloud_detection, **options)



import os
import logging
import traceback
import glob
import numpy as np
import ds_format as ds
import aquarius_time as aq
from alcf.lidars import LIDARS
from alcf.algorithms.calibration import CALIBRATION
from alcf.algorithms.noise_removal import NOISE_REMOVAL
from alcf.algorithms.cloud_detection import CLOUD_DETECTION
from alcf.algorithms.cloud_base_detection import CLOUD_BASE_DETECTION
from alcf.algorithms import tsample, zsample, output_sample, lidar_ratio
from alcf.algorithms import couple as couple_mod
from alcf import misc
import pst

VARIABLES = [
	'backscatter',
	'backscatter_mol',
	'backscatter_sd',
	'time',
	'time_bnds',
	'zfull',
	'altitude',
	'lon',
	'lat',
]

def read_calibration_file(filename):
	with open(filename, 'rb') as f:
		return pst.decode(f.read())

def run(type_, input_, output,
	altitude=None,
	tres=300,
	time=None,
	tshift=0.,
	zres=50,
	zlim=[0., 15000.],
	cloud_detection='default',
	cloud_base_detection='default',
	noise_removal='default',
	calibration='default',
	output_sampling=86400,
	overlap_file=None,
	calibration_file=None,
	couple=None,
	fix_cl_range=False,
	cl_crit_range=6000,
	lat=None,
	lon=None,
	r=False,
	keep_vars=[],
	align_output=True,
	**options
):
	"""
alcf-lidar -- Process lidar data.
==========

Synopsis
--------

    alcf lidar <type> [<options>] [<algorithm_options>] [--] <lidar> <output>

Description
-----------

The processing is done in the following order:

- noise removal
- calibration
- time resampling
- height resampling
- cloud detection
- cloud base detection

Arguments following `--` are treated as literal strings. Use this delimiter if the input or output file names might otherwise be interpreted as non-strings, e.g. purely numerical file names.

Arguments
---------

- `type`: Lidar type (see Types below).
- `lidar`: Input lidar data directory or filename. If a directory, only `.nc` files in the directory are processed. If the option `-r` is supplied, the directory is processed recursively.
- `output`: Output filename or directory.
- `options`: See Options below.
- `algorithm_options`: See Algorithm options below.

Types
-----

- `blview`: Vaisala BL-VIEW L2 product.
- `chm15k`: Lufft CHM 15k.
- `cl31`: Vaisala CL31.
- `cl51`: Vaisala CL51.
- `cl61`: Vaisala CL61.
- `cosp`: COSP simulated lidar.
- `default`: The same format as the output of `alcf lidar`.
- `minimpl`: Sigma Space MiniMPL (converted via SigmaMPL).
- `mpl`: Sigma Space MPL (converted via SigmaMPL).
- `mpl2nc`: Sigma Space MPL and MiniMPL (converted via mpl2nc).

Options
-------

- `align_output: <value>`: Align output time periods to the nearest multiple of output_sampling. Default: `true`.
- `altitude: <altitude>`: Altitude of the instrument (m). Default: Taken from lidar data or `0` if not available.
- `bsd: <value>`: Assume a given standard deviation of backscatter noise when detecting clouds or `none` to use the value calculated by the noise removal algorithm from observed backscatter if available (m^-1.sr^-1). The value applies at height `bsd_z` and is range-scaled for other heights. A suitable value can be taken from a plot generated by `alcf plot backscatter_sd_hist`. Default: `none`.
- `bsd_z: <value>`: Height at which `bsd` applies (m). Default: `8000`.
- `calibration: <algorithm>`: Backscatter calibration algorithm. Available algorithms: `default`, `none`. Default: `default`.
- `couple: <directory>`: Couple to other lidar data. Default: `none`.
- `cl_crit_range: <range>`: Critical range for the `fix_cl_range` option (m). Default: 6000.
- `cloud_detection: <algorithm>`: Cloud detection algorithm. Available algorithms: `default`, `none`. Default: `default`.
- `cloud_base_detection: <algorithm>`: Cloud base detection algorithm. Available algorithms: `default`, `none`. Default: `default`.
- `--fix_cl_range`: Fix CL31/CL51 range correction (if `noise_h2` firmware option if off). The critical range is taken from `cl_crit_range`.
- `keep_vars: { <var>... }`: Keep the listed input variables. The variable must be numerical and have a time dimension. The variable is resampled in the same way as backscatter along their time and level dimensions, its name is prefixed with `input_`, and its type is changed to float64. Default: `{ }`.
- `lat: <lat>`: Latitude of the instrument (degrees North). Default: Taken from lidar data or `none` if not available.
- `lon: <lon>`: Longitude of the instrument (degrees East). Default: Taken from lidar data or `none` if not available.
- `noise_removal: <algorithm>`: Noise removal algorithm. Available algorithms: `default`, `none`.  Default: `default`.
- `output_sampling: <period>`: Output sampling period (seconds). Default: `86400` (24 hours).
- `-r`: Process the input directory recursively.
- `time: { <low> <high> }`: Time limits (see Time format below). Default: `none`.
- `tres: <tres>`: Time resolution (seconds). Default: `300` (5 min).
- `tshift: <tshift>`: Time shift (seconds). Default: `0`.
- `zlim: { <low> <high> }`: Height limits (m). Default: `{ 0 15000 }`.
- `zres: <zres>`: Height resolution (m). Default: `50`.

Cloud detection options
-----------------------

- `default`: Cloud detection based on backscatter threshold.
- `none`: Disable cloud detection.

Cloud detection default options
-------------------------------

- `cloud_nsd: <n>`: Number of noise standard deviations to subtract. Default: `5`.
- `cloud_threshold: <threshold>`: Cloud detection threshold (m^-1.sr^-1). Default: `2e-6`.
- `cloud_threshold_exp: { <x> <y> <h> }`: Cloud detection threshold exponentially decaying with height (sr^-1.m^-1). If not `none`, this supersedes `cloud_threshold`. The threshold is `<x>` at surface level, decaying exponentially to `<y>` at infinite height with half-height `<h>`. Default: `none`.

Cloud base detection options
----------------------------

- `default`: Cloud base detection based cloud mask produced by the cloud detection algorithm.
- `none`: Disable cloud base detection.

Calibration options
-------------------

- `default`: Multiply backscatter by a calibration coefficient.
- `none`: Disable calibration.

Calibration default options
---------------------------

- `calibration_file: <file>`: Calibration file.

Noise removal options
---------------------

- `default`: Noise removal based on noise distribution on the highest level.
- `none`: Disable noise removal.

Noise removal default options
-----------------------------

- `noise_removal_sampling: <period>`: Sampling period for noise removal (seconds). Default: 300.
- `near_noise: { <scale> <range> }` : Assume additional exponentially-decaying near-range noise. The first argument is the value at zero range (sr^-1.m^-1). The second argument is range at which the function decays to a half (m). Default: `{ 0 0 }`.

Time format
-----------

`YYYY-MM-DD[THH:MM[:SS]]`, where `YYYY` is year, `MM` is month, `DD` is day, `HH` is hour, `MM` is minute, `SS` is second. Example: `2000-01-01T00:00:00`.

Examples
--------

Process Vaisala CL51 data in `cl51_nc` and store the output in `cl51_alcf_lidar`, assuming instrument altitude of 100 m above sea level.

    alcf lidar cl51 cl51_nc cl51_alcf_lidar altitude: 100
	"""
	# if time is not None:
	# 	start, end = misc.parse_time(time)

	lidar = LIDARS.get(type_)
	if lidar is None:
		raise ValueError('Invalid type: %s' % type_)

	time1 = None
	if time is not None:
		time1 = [None, None]
		for i in 0, 1:
			time1[i] = aq.from_iso(time[i])
			if time1[i] is None:
				raise ValueError('Invalid time format: %s' % time[i])

	noise_removal_mod = None
	calibration_mod = None
	cloud_detection_mod = None
	cloud_base_detection_mod = None

	if type_ not in ('default', 'cosp') and noise_removal is not None:
		noise_removal_mod = NOISE_REMOVAL.get(noise_removal)
		if noise_removal_mod is None:
			raise ValueError('Invalid noise removal algorithm: %s' % noise_removal)

	if calibration is not None:
		calibration_mod = CALIBRATION.get(calibration)
		if calibration_mod is None:
			raise ValueError('Invalid calibration algorithm: %s' % calibration)

	if cloud_detection is not None:
		cloud_detection_mod = CLOUD_DETECTION.get(cloud_detection)
		if cloud_detection_mod is None:
			raise ValueError('Invalid cloud detection algorithm: %s' % cloud_detection)

	if cloud_base_detection is not None:
		cloud_base_detection_mod = CLOUD_BASE_DETECTION.get(cloud_base_detection)
		if cloud_base_detection_mod is None:
			raise ValueError('Invalid cloud base detection algorithm: %s' % cloud_base_detection)

	if calibration_file is not None:
		c = read_calibration_file(calibration_file)
		calibration_coeff = c[b'calibration_coeff']/lidar.CALIBRATION_COEFF
	else:
		calibration_coeff = 1.

	def write(d, output):
		if len(d['time']) == 0:
			return
		t1 = d['time_bnds'][0,0]
		t1 = np.round(t1*86400.)/86400.
		filename = os.path.join(output, '%s.nc' % aq.to_iso(t1).replace(':', ''))
		ds.write(filename, d)
		print('-> %s' % filename)
		return []

	def output_stream(dd, state, output_sampling=None, **options):
		return misc.stream(dd, state, write, output=output)

	def preprocess(d, tshift=None):
		if tshift is not None:
			d['time'] += tshift/86400.
			d['time_bnds'] += tshift/86400.
		return d

	def process(dd, state, **options):
		state['preprocess'] = state.get('preprocess', {})
		state['noise_removal'] = state.get('noise_removal', {})
		state['calibration'] = state.get('calibration', {})
		state['tsample'] = state.get('tsample', {})
		state['zsample'] = state.get('zsample', {})
		state['output_sample'] = state.get('output_sample', {})
		state['output_sample_2'] = state.get('output_sample_2', {})
		state['cloud_detection'] = state.get('cloud_detection', {})
		state['cloud_base_detection'] = state.get('cloud_base_detection', {})
		state['lidar_ratio'] = state.get('lidar_ratio', {})
		state['output'] = state.get('output', {})
		state['couple'] = state.get('couple', {})
		dd = misc.stream(dd, state['preprocess'], preprocess, tshift=tshift)
		if couple is not None:
			dd = couple_mod.stream(dd, state['couple'], couple)
		if noise_removal_mod is not None:
			dd = noise_removal_mod.stream(dd, state['noise_removal'],
				align=align_output,
				**options
			)
		if calibration_mod is not None:
			dd = calibration_mod.stream(dd, state['calibration'], **options)
		if zres is not None or zlim is not None:
			dd = zsample.stream(dd, state['zsample'], zres=zres, zlim=zlim)
		if tres is not None or time is not None:
			dd = tsample.stream(dd, state['tsample'],
				tres=tres/86400.,
				align=align_output
			)
		if output_sampling is not None:
			dd = output_sample.stream(dd, state['output_sample'],
				tres=tres/86400.,
				output_sampling=output_sampling/86400.,
				align=align_output,
			)
			dd = misc.aggregate(
				dd,
				state['output_sample_2'],
				output_sampling/86400.,
				align=align_output,
			)
		#for d in dd:
		#	if d is not None:
		#		print(aq.to_iso(d['time_bnds'][0,0]), aq.to_iso(d['time_bnds'][-1,1]))
		if cloud_detection_mod is not None:
			dd = cloud_detection_mod.stream(dd, state['cloud_detection'],
				**options
			)
		if cloud_base_detection_mod is not None:
			dd = cloud_base_detection_mod.stream(
				dd,
				state['cloud_base_detection'],
				**options
			)
		dd = lidar_ratio.stream(dd, state['lidar_ratio'])
		dd = output_stream(dd, state['output'])
		return dd

	options['output'] = output
	options['calibration_coeff'] = calibration_coeff

	state = {}
	if os.path.isdir(input_):
		pattern = '**/*.nc' if r else '*.nc'
		files = sorted(glob.glob(os.path.join(
			glob.escape(input_),
			pattern
		), recursive=r))
		for file_ in files:
			if not os.path.isfile(file_):
				continue
			print('<- %s' % file_)
			try:
				d = lidar.read(file_, VARIABLES,
					altitude=altitude,
					lon=lon,
					lat=lat,
					fix_cl_range=fix_cl_range,
					cl_crit_range=cl_crit_range,
					tlim=time1,
					keep_vars=keep_vars,
				)
				if d is None: continue
				dd = process([d], state, **options)
			except SystemExit:
				raise
			except SystemError:
				raise
			except:
				logging.warning(traceback.format_exc())
		dd = process([None], state, **options)
	else:
		print('<- %s' % input_)
		d = lidar.read(input_, VARIABLES,
			altitude=altitude,
			lon=lon,
			lat=lat,
			fix_cl_range=fix_cl_range,
			cl_crit_range=cl_crit_range,
			tlim=time1,
			keep_vars=keep_vars,
		)
		dd = process([d, None], state, **options)



import os
from alcf.cmds import model, simulate, lidar, stats, plot
from alcf.cmds.auto_cmds import lidar as auto_lidar

STEPS = ['model', 'simulate'] + auto_lidar.STEPS

def run(model_type, lidar_type, input_, output, *args, skip=None, **kwargs):
	model_dir = os.path.join(output, 'model')
	simulate_dir = os.path.join(output, 'simulate')

	if skip is not None:
		try: i = STEPS.index(skip)
		except ValueError:
			raise ValueError('Invalid step "%s"' % skip)
	else:
		i = -1

	print('-> %s' % output)
	try: os.mkdir(output)
	except OSError: pass

	if i < STEPS.index('model'):
		print('-> %s' % model_dir)
		try: os.mkdir(model_dir)
		except OSError: pass
		print('! alcf model')
		model.run(model_type, input_, model_dir, *args, **kwargs)
	if i < STEPS.index('simulate'):
		print('-> %s' % simulate_dir)
		try: os.mkdir(simulate_dir)
		except OSError: pass
		print('! alcf simulate')
		simulate.run(lidar_type, model_dir, simulate_dir)

	lidar_skip = skip if i > STEPS.index('simulate') else None
	auto_lidar.run('cosp', simulate_dir, output, *args,
		skip=lidar_skip,
		**kwargs
	)



# -*- coding: utf-8 -*-

import os
import sys
import logging
import traceback
import copy
import numpy as np
import datetime as dt
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.dates import date2num, AutoDateFormatter, AutoDateLocator
from matplotlib.colors import ListedColormap, Normalize, BoundaryNorm, LogNorm
import matplotlib.lines as mlines
import aquarius_time as aq
import ds_format as ds
from alcf import misc, algorithms
from alcf.lidars import LIDARS

COLORS = [
	'#0084c8',
	'#dc0000',
	'#009100',
	'#ffc022',
	'#ba00ff',
]

LINESTYLE = 'solid'

VARIABLES = [
	'time',
	'time_bnds',
	'backscatter',
	'backscatter_sd',
	'backscatter_mol',
	'backscatter_sd_full',
	'backscatter_sd_hist',
	'zfull',
	'lr',
	'cbh',
	'cl',
	'clt',
	'n',
	'cloud_mask',
	'backscatter_hist',
	'backscatter_full',
	'altitude',
	'clw',
	'cli',
]

def time_to_dt(time):
	time_dt = aq.to_datetime(time)
	for i, t in enumerate(time_dt):
		if t.microsecond >= 0.5e6:
			t += dt.timedelta(microseconds=(1e6 - t.microsecond))
		else:
			t -= dt.timedelta(microseconds=t.microsecond)
		time_dt[i] = t
	return time_dt

def plot_legend(*args, theme='light', **kwargs):
	legend = plt.legend(*args, fontsize=8, **kwargs)
	f = legend.get_frame()
	f.set_facecolor('k' if theme == 'light' else 'white')
	f.set_linewidth(0)
	f.set_alpha(0.1 if theme == 'light' else 0.9)

def plot_profile(plot_type, d,
	cax=None,
	subcolumn=0,
	sigma=5,
	remove_bmol=True,
	vlim=None,
	vlog=None,
	zres=50,
	zlim=None,
	render='antialiased',
	**opts
):
	if plot_type == 'backscatter':
		cmap = copy.copy(mpl.cm.get_cmap('viridis'))
		under = '#222222'
		label = 'Att. vol. backscattering coef. (×10$^{-6}$ m$^{-1}$sr$^{-1}$)'
		if vlim is None:
			vlim = [0.1, 200]
		if vlog is None:
			vlog = True
		if len(d['backscatter'].shape) == 3:
			b = d['backscatter'][:,:,subcolumn]
			cloud_mask = d['cloud_mask'][:,:,subcolumn]
			bsd = d['backscatter_sd'][:,:,subcolumn] if 'backscatter_sd' in d \
				else np.zeros(b.shape, dtype=np.float64)
		else:
			b = d['backscatter']
			cloud_mask = d['cloud_mask']
			bsd = d['backscatter_sd'] if 'backscatter_sd' in d \
				else np.zeros(b.shape, dtype=np.float64)
		if sigma > 0:
			b -= sigma*bsd
		if remove_bmol and 'backscatter_mol' in d:
			bmol = d['backscatter_mol']
			mask = ~np.isnan(bmol)
			b[mask] -= bmol[mask]
		x = b*1e6
		x[x <= 0.] = 0.5*vlim[0] # A value below the limit.
		zfull = d['zfull']
	elif plot_type in ('clw', 'cli', 'cl'):
		cmap = copy.copy(mpl.cm.get_cmap({
			'clw': 'Reds',
			'cli': 'Blues',
			'cl': 'Greys_r',
		}[plot_type]))
		under = {
			'clw': 'white',
			'cli': 'white',
			'cl': 'k',
		}[plot_type]
		label = {
			'clw': 'Cloud water (g/kg)',
			'cli': 'Cloud ice (g/kg)',
			'cl': 'Cloud fraction (%)',
		}[plot_type]
		if vlim is None:
			vlim = {
				'clw': [1e-3, 1],
				'cli': [1e-3, 1],
				'cl': [0, 100],
			}[plot_type]
		if vlog is None:
			vlog = {
				'clw': True,
				'cli': True,
				'cl': False
			}[plot_type]
		x = d[plot_type]
		if plot_type in ('clw', 'cli', 'clw+cli'):
			x *= 1e3
		if x.shape == 3:
			x = x[:,:,subcolumn]
		if zlim is None:
			zlim = [np.min(d['zfull']), np.max(d['zfull'])]
		zhalf = np.arange(zlim[0], zlim[1] + zres, zres)
		zfull = 0.5*(zhalf[1:] + zhalf[:-1])
		xp = np.full((x.shape[0], len(zfull)), np.nan, np.float64)
		for i in range(xp.shape[0]):
			zhalfi = misc.half(d['zfull'][i,:])
			xp[i,:] = algorithms.interp(
				zhalfi, x[i,:], zhalf
			)
		x = xp
	else:
		raise ValueError('Invalid plot type "%s"' % plot_type)

	if vlog:
		norm = LogNorm(vlim[0], vlim[1])
	else:
		norm = Normalize(vlim[0], vlim[1])

	time = d['time']
	time_dt = time_to_dt(time)

	t1 = time_dt[0] - 0.5*(time_dt[1] - time_dt[0])
	t2 = time_dt[-1] + 0.5*(time_dt[-1] - time_dt[-2])
	z1 = zfull[0] - 0.5*(zfull[1] - zfull[0])
	z2 = zfull[-1] + 0.5*(zfull[-1] - zfull[-2])

	try:
		interpolation = {
			'antialiased': 'antialiased',
			'standard': 'none',
		}[render]
	except KeyError:
		raise ValueError('Invalid rendering method "%s"' % render) from None

	im = plt.imshow(x.T,
		extent=(date2num(t1), date2num(t2), z1*1e-3, z2*1e-3),
		aspect='auto',
		origin='lower',
		norm=norm,
		cmap=cmap,
		interpolation=interpolation,
	)
	im.cmap.set_under(under)

	cb = plt.colorbar(
		cax=cax,
		label=label,
		pad=0.03,
		fraction=0.05,
		extend='both',
	)

	if zlim is not None:
		plt.ylim(zlim[0]*1e-3, zlim[1]*1e-3)

	plt.xlabel('Time (UTC)')
	plt.ylabel('Height (km)')

	def format_func(t, p):
		return mpl.dates.num2date(t).strftime('%d/%m\n%H:%M')

	formatter = plt.FuncFormatter(format_func)
	locator = AutoDateLocator()
	plt.gca().xaxis.set_major_formatter(formatter)
	plt.gca().xaxis.set_major_locator(locator)

	if plot_type == 'backscatter' and opts.get('cloud_mask'):
		time_dt_exp = np.array([t1] + time_dt.tolist() + [t2])
		cloud_mask_exp = np.concatenate((
			cloud_mask[0:1,:],
			cloud_mask,
			cloud_mask[-2:-1,:],
		))
		cf = plt.contour(
			date2num(time_dt_exp),
			d['zfull']*1e-3,
			cloud_mask_exp.T,
			colors='red',
			linewidths=1,
			linestyles='dashed',
			levels=[-1., 0.5, 2.],
		)
		plot_legend(
			handles=[mlines.Line2D([], [],
				color='red', linestyle='dashed', label='Cloud mask'
			)],
			theme='dark'
		)

	if 'altitude' in d:
		plt.plot(date2num(time_dt), d['altitude']*1e-3, color='red', lw=0.5)

def plot_lr(d, subcolumn=0, **opts):
	lr = d['lr'][:,subcolumn] if d['lr'].ndim == 2 else d['lr'][:]
	plt.plot(d['time'], lr, lw=0.7, color='#0087ed')
	locator = AutoDateLocator()
	plt.gca().xaxis.set_major_locator(locator)
	plt.grid(lw=0.1, color='black')
	def f(x, pos):
		return aq.to_datetime(x).strftime('%d/%m\n%H:%M')
	formatter = plt.FuncFormatter(f)
	plt.gca().xaxis.set_major_formatter(formatter)
	plt.xlim(np.min(d['time']), np.max(d['time']))
	plt.ylim(0, 50)
	plt.ylabel('Eff. lidar ratio (sr)')
	plt.xlabel('Time (UTC)')

def plot_cloud_dist(plot_type, dd,
	colors=COLORS,
	linestyle=LINESTYLE,
	lw=None,
	labels=None,
	subcolumn=0,
	xlim=[0., 100.],
	zlim=[0., 15000],
	**kwargs
):
	var = {
		'cbh': 'cbh',
		'cloud_occurrence': 'cl',
	}[plot_type]
	for i, d in enumerate(dd):
		zfull = d['zfull']
		x = d[var][:,subcolumn] \
			if len(d[var].shape) == 2 \
			else d[var]
		label = (labels[i] if labels is not None else '')
		if plot_type == 'cloud_occurrence':
			clt = d['clt'][subcolumn] \
				if len(d['clt'].shape) == 1 \
				else d['clt']
			label += ' | CF: %d%%' % clt
		plt.plot(x, 1e-3*zfull,
			color=colors[i],
			linestyle=(linestyle[i] if type(linestyle) is list else linestyle),
			lw=lw,
			label=label,
		)
	plt.xlim(xlim[0], xlim[1])
	plt.ylim(zlim[0]*1e-3, zlim[1]*1e-3)
	plt.xlabel({
		'cbh': 'Cloud base height distribution (%)',
		'cloud_occurrence': 'Cloud occurrence (%)',
	}[plot_type])
	plt.ylabel('Height (km)')

	if labels is not None:
		plot_legend()

def plot_backscatter_hist(d,
	subcolumn=0,
	xlim=None,
	zlim=None,
	vlim=None,
	vlog=False,
	**kwargs
):
	if vlim is None:
		vlim = [
			np.min(d['backscatter_hist'])*1e2,
			np.max(d['backscatter_hist'])*1e2
		]

	if vlog is False:
		norm = Normalize(vlim[0], vlim[1])
	else:
		if vlim[0] <= 0: vlim[0] = 1e-3
		norm = LogNorm(vlim[0], vlim[1])

	if xlim is None:
		xlim = [
			(1.5*d['backscatter_full'][0] - 0.5*d['backscatter_full'][1])*1e6,
			(1.5*d['backscatter_full'][-1] - 0.5*d['backscatter_full'][-2])*1e6,
		]

	under = '#222222'
	cmap = copy.copy(mpl.cm.get_cmap('viridis'))
	cmap.set_under(under)
	im = plt.imshow(
		d['backscatter_hist'].T*1e2
			if len(d['backscatter_hist'].shape) == 2
			else d['backscatter_hist'][:,:,subcolumn].T*1e2,
		origin='lower',
		aspect='auto',
		extent=(
			(1.5*d['backscatter_full'][0] - 0.5*d['backscatter_full'][1])*1e6,
			(1.5*d['backscatter_full'][-1] - 0.5*d['backscatter_full'][-2])*1e6,
			(1.5*d['zfull'][0] - 0.5*d['zfull'][1])*1e-3,
			(1.5*d['zfull'][-1] - 0.5*d['zfull'][-2])*1e-3,
		),
		norm=norm,
		cmap=cmap,
	)
	plt.gca().set_facecolor(under)
	plt.gca().set_box_aspect(1)
	cax = plt.gca().inset_axes([1.03, 0, 0.03, 1],
		transform=plt.gca().transAxes)
	plt.gcf().colorbar(im,
		cax=cax,
		extend='both',
		label='Occurrence (%)',
	)
	if xlim is not None:
		plt.xlim(xlim)
	if zlim is not None:
		plt.ylim(zlim[0]*1e-3, zlim[1]*1e-3)
	plt.xlabel('Attenuated volume backscattering coefficient (×10$^{-6}$ m$^{-1}$sr$^{-1}$)')
	plt.ylabel('Height (km)')
	plt.axvline(0, lw=0.3, linestyle='dashed', color='k')

def plot_backscatter_sd_hist(dd,
	labels=None,
	xlim=None,
	zlim=None,
	colors=COLORS,
	linestyle=LINESTYLE,
	**kwargs
):
	for i, d in enumerate(dd):
		plt.plot(d['backscatter_sd_full']*1e6, d['backscatter_sd_hist'],
			lw=1,
			color=colors[i],
			linestyle=(linestyle[i] if type(linestyle) is list else linestyle),
			label=(labels[i] if labels is not None else None),
		)
	plt.gca().set_yscale('log')
	plt.gca().set_xscale('log')
	if xlim is not None:
		plt.xlim(xlim)
	if zlim is not None:
		plt.ylim(zlim)
	plt.xlabel('Attenuated volume backscattering coefficient (×10$^{-6}$ m$^{-1}$sr$^{-1}$)')
	plt.ylabel('Occurrence (%)')
	if labels is not None:
		plot_legend()

def plot(plot_type, d, output,
	width=None,
	height=None,
	lr=False,
	grid=False,
	dpi=300,
	title=None,
	**kwargs
):
	fontsdir = os.path.join(os.path.dirname(__file__), '../fonts')
	for font in mpl.font_manager.findSystemFonts(fontsdir):
		mpl.font_manager.fontManager.addfont(font)
	mpl.rcParams['font.family'] = 'Public Sans'
	mpl.rcParams['axes.linewidth'] = 0.5
	mpl.rcParams['xtick.major.size'] = 3
	mpl.rcParams['xtick.major.width'] = 0.5
	mpl.rcParams['ytick.major.size'] = 3
	mpl.rcParams['ytick.major.width'] = 0.5

	if width is not None and height is not None:
		fig = plt.figure(figsize=[width, height])

	if plot_type == 'backscatter':
		if lr:
			gs = GridSpec(2, 2,
				width_ratios=[0.985, 0.015],
				height_ratios=[0.7, 0.3],
				hspace=0.4,
				wspace=0.05,
			)
			cax = plt.subplot(gs[1])
			ax = plt.subplot(gs[0])
		else:
			gs = GridSpec(1, 2,
				width_ratios=[0.985, 0.015],
				wspace=0.05,
			)
			cax = plt.subplot(gs[1])
			ax = plt.subplot(gs[0])
		plot_profile(plot_type, d, cax, **kwargs)
		if lr:
			plt.subplot(gs[2])
			plot_lr(d, **kwargs)
	elif plot_type in ('clw', 'cli', 'cl'):
		gs = GridSpec(1, 2,
			width_ratios=[0.985, 0.015],
			wspace=0.05,
		)
		cax = plt.subplot(gs[1])
		ax = plt.subplot(gs[0])
		plot_profile(plot_type, d, cax, **kwargs)
	elif plot_type == 'clw+cli':
		r = 10./(plt.gcf().get_figwidth())*0.045
		gs = GridSpec(1, 4,
			width_ratios=[max(0.015, 1. - 2*0.015 - r), 0.015, r, 0.015],
			wspace=0.15,
			figure=plt.gcf(),
		)
		cax1 = plt.subplot(gs[1])
		cax2 = plt.subplot(gs[3])
		ax = plt.subplot(gs[0])
		plot_profile('clw', d, cax1, alpha=0.5, **kwargs)
		plot_profile('cli', d, cax2, alpha=0.5, **kwargs)
	elif plot_type in ('cbh', 'cloud_occurrence'):
		plot_cloud_dist(plot_type, d, **kwargs)
	elif plot_type == 'backscatter_hist':
		plot_backscatter_hist(d, **kwargs)
	elif plot_type == 'backscatter_sd_hist':
		plot_backscatter_sd_hist(d, **kwargs)
	else:
		raise ValueError('Invalid plot type "%s"' % plot_type)

	if grid:
		plt.grid(lw=0.1, color='k', alpha=1)

	if title is not None:
		plt.title(title)

	plt.savefig(output, bbox_inches='tight', dpi=dpi)

def run(plot_type, *args,
	lr=True,
	subcolumn=0,
	width=None,
	height=None,
	dpi=300,
	grid=False,
	colors=COLORS,
	linestyle=LINESTYLE,
	lw=1,
	labels=None,
	xlim=None,
	zlim=None,
	vlim=None,
	vlog=None,
	sigma=5.,
	remove_bmol=True,
	cloud_mask=True,
	title=None,
	zres=50,
	render='antialiased',
	**kwargs
):
	'''
alcf-plot -- Plot lidar data.
=========

Synopsis
--------

    alcf plot <plot_type> [<options>] [<plot_options>] [--] <input> <output>

Description
-----------

Arguments following `--` are treated as literal strings. Use this delimiter if the input or output file names might otherwise be interpreted as non-strings, e.g. purely numerical file names.

Arguments
---------

- `plot_type`: Plot type (see Plot types below).
- `input`: Input filename or directory.
- `output`: Output filename or directory.
- `options`: See Options below.
- `plot_options`: Plot type specific options. See Plot options below.

Plot types
----------

- `backscatter`: Plot backscatter from `alcf lidar` output on time-height axes.
- `backscatter_hist`: Plot backscatter histogram from `alcf stats` output on backscatter-height axes.
- `backscatter_sd_hist`: Plot backscatter standard deviation histogram from `alcf stats` output.
- `cbh`: Plot cloud base height distribution from `alcf stats` output.
- `cl`: Plot model cloud area fraction from `alcf lidar` output on time-height axes.
- `cli`: Plot model mass fraction of cloud ice from `alcf lidar` output on time-height axes.
- `cloud_occurrence`: Plot cloud occurrence by height from `alcf stats` output.
- `clw`: Plot model mass fraction of cloud liquid water from `alcf lidar` output on time-height axes.
- `clw+cli`: Plot model mass fraction of cloud liquid water and ice from `alcf lidar` output on time-height axes.

General options
---------------

- `dpi: <value>`: Resolution in dots per inch (DPI). Default: `300`.
- `--grid`: Plot grid.
- `height: <value>`: Plot height (inches). Default: `5` if `plot_type` is `cloud_occurrence` or `backscatter_hist` else `4`.
- `subcolumn: <value>`: Model subcolumn to plot. Default: `0`.
- `title: <value>`: Plot title.
- `width: <value>`: Plot width (inches). Default: `5` if `plot_type` is `cloud_occurrence` or `backscatter_hist` else `10`.
- `render: <value>`: Render profiles anti-aliased (`antialiased`) or standard (`standard`). Standard is more suitable for short time intervals. Default: `antialiased`.

backscatter options
-------------------

- `lr: <value>`: Plot effective lidar ratio. Default: `true`.
- `cloud_mask: <value>`: Plot cloud mask. Default: `true`.
- `remove_bmol: <value>`: Remove molecular backscatter (observed data have to be coupled with model data via the `couple` option of `alcf lidar`). Default: `true`.
- `sigma: <value>`: Remove of number of standard deviations of backscatter from the mean backscatter (real). Default: `5`.
- `vlim: { <min> <max }`: Value limits (10^-6 m-1.sr-1). Default: `{ 0.1 200 }`.
- `vlog: <value>`: Plot values on logarithmic scale: `true` of `false`. Default: `true`.

backscatter_hist options
------------------------

- `vlim: { <min> <max> }`: Value limits (%) or `none` for auto. If `none` and `vlog` is `none`, `min` is set to 1e-3 if less or equal to zero. Default: `none`.
- `--vlog`: Use logarithmic scale for values.
- `xlim: { <min> <max> }`: x axis limits (10^-6 m-1.sr-1) or `none` for automatic. Default: `none`.
- `zlim: { <min> <max> }`: z axis limits (m) or `none` for automatic. Default: `none`.

backscatter_sd_hist options
---------------------------

- `xlim: { <min> <max> }`: x axis limits (10^-6 m-1.sr-1) or `none` for automatic. Default: `none`.
- `zlim: { <min> <max> }`: z axis limits (%) or `none` for automatic. Default: `none`.

cl options
----------

- `vlim: { <min> <max> }`: Value limits (%). Default: `{ 0 100 }`.
- `vlog: <value>`: Plot values on logarithmic scale: `true` of `false`. Default: `false`.
- `zres: <zres>`: Height resolution (m). Default: `50`.

cli, clw, and clw+cli options
-----------------------------

- `vlim: { <min> <max> }`: Value limits (g/kg). Default: `{ 1e-3 1 }`.
- `vlog: <value>`: Plot values on logarithmic scale: `true` of `false`. Default: `true`.
- `zres: <zres>`: Height resolution (m). Default: `50`.

cbh and cloud_occurrence options
--------------------------------

- `colors: { <value>... }`: Line colors. Default: `{ #0084c8 #dc0000 #009100 #ffc022 #ba00ff }`
- `linestyle: { <value> ... }`: Line style (`solid`, `dashed`, `dotted`). Default: `solid`.
- `labels: { <value>... }`: Line labels. Default: `none`.
- `lw: <value>`: Line width. Default: `1`.
- `xlim: { <min> <max> }`: x axis limits (%). Default: `{ 0 100 }`.
- `zlim: { <min> <max> }`: z axis limits (m). Default: `{ 0 15 }`.

Examples
--------

Plot backscatter from processed Vaisala CL51 data in `alcf_cl51_lidar` and store the output in the directory `alcf_cl51_backscatter`.

    alcf plot backscatter alcf_cl51_lidar alcf_cl51_backscatter
	'''
	input_ = args[:-1]
	output = args[-1]

	if plot_type in ('backscatter_hist', 'backscatter_sd_hist', 'cbh', 'cloud_occurrence'):
		width = width if width is not None else 5
		height = height if height is not None else 5
	else:
		width = width if width is not None else 10
		height = height if height is not None else 6

	opts = {
		'lr': lr,
		'subcolumn': subcolumn,
		'grid': grid,
		'colors': colors,
		'linestyle': linestyle,
		'lw': lw,
		'labels': labels,
		'sigma': sigma,
		'remove_bmol': remove_bmol,
		'title': title,
		'cloud_mask': cloud_mask,
		'width': width,
		'height': height,
		'render': render,
	}

	if xlim is not None: opts['xlim'] = xlim
	if zlim is not None: opts['zlim'] = zlim
	if vlim is not None: opts['vlim'] = vlim
	if vlog is not None: opts['vlog'] = vlog
	if zres is not None: opts['zres'] = zres

	state = {}
	if plot_type in ('cbh', 'cloud_occurrence', 'backscatter_sd_hist'):
		dd = []
		for file in input_:
			print('<- %s' % file)
			dd += [ds.read(file, VARIABLES)]
		plot(plot_type, dd, output, **opts)
		print('-> %s' % output)
	elif plot_type == 'backscatter_hist':
		print('<- %s' % input_[0])
		d = ds.read(input_[0], VARIABLES)
		plot(plot_type, d, output, **opts)
		print('-> %s' % output)
	elif plot_type in ('backscatter', 'clw', 'cli', 'clw+cli', 'cl'):
		for input1 in input_:
			if os.path.isdir(input1):
				for file_ in sorted(os.listdir(input1)):
					filename = os.path.join(input1, file_)
					output_filename = os.path.join(
						output,
						os.path.splitext(file_)[0] + '.png'
					)
					try:
						print('<- %s' % filename)
						d = ds.read(filename, VARIABLES)
					except SystemExit:
						raise
					except SystemError:
						raise
					except:
						logging.warning(traceback.format_exc())
					try:
						plot(plot_type, d, output_filename, **opts)
						print('-> %s' % output_filename)
					except SystemExit:
						raise
					except SystemError:
						sys.exit(1)
					except:
						logging.warning(traceback.format_exc())
					finally:
						plt.close()
			else:
				print('<- %s' % input1)
				d = ds.read(input1, VARIABLES)
				try:
					plot(plot_type, d, output, **opts)
				except SystemExit:
					raise
				except SystemError:
					sys.exit(1)
	else:
		raise ValueError('Invalid plot type "%s"' % plot_type)



import numpy as np
from alcf import misc

def noise_removal(d, **options):
	b = d['backscatter']
	zfull = d['zfull']
	bt = b[:,-1]
	n, m = b.shape
	b2 = np.zeros((n, m), np.float64)
	b_sd = np.zeros((n, m), np.float64)
	w = d['time_bnds'][:,1] - d['time_bnds'][:,0]
	noise_m = np.average(bt, weights=w)
	noise_sd = np.sqrt(np.cov(bt, aweights=w)) if len(bt) > 1 else 0.
	for i in range(n):
		c = (1.0*zfull[i,:]/zfull[i,-1])**2
		b2[i,:] = b[i,:] - noise_m*c
		b_sd[i,:] = noise_sd*c
	nn_scale, nn_range = options.get('near_noise', [0, 0])
	if nn_scale > 0 and nn_range > 0:
		nn_lambda = np.log(2)/nn_range
		for i in range(n):
			near_noise = nn_scale*np.exp(-zfull[i,:]*nn_lambda)
			b_sd[i,:] += near_noise
	elif nn_scale == 0:
		pass
	else:
		raise ValueError('Near-range noise scale must non-negative and range must be positive')
	d['backscatter'] = b2
	d['backscatter_sd'] = b_sd
	d['.']['backscatter_sd'] = {
		'.dims': ['time', 'level'],
		'long_name': 'total attenuated volume backscattering coefficient standard deviation',
		'units': 'm-1 sr-1',
	}

def stream(dd, state, noise_removal_sampling=300, align=True, **options):
	state['aggregate_state'] = state.get('aggregate_state', {})
	dd = misc.aggregate(
		dd,
		state['aggregate_state'],
		noise_removal_sampling/60./60./24.,
		align=align,
	)
	return misc.stream(dd, state, noise_removal, **options)
